use super::{MinimizerAgent, RandomAgent};
use crate::{
	board::{Board, PlayerAction},
	util::Symb,
};

/// A simple "operator diffusion" MINIMIZER agent.
///
/// Tries to keep operators as far apart as possible, denying large numbers.
pub struct DiffuseAgent {}

impl DiffuseAgent {
	/// Place a symbol on the board.
	/// Assumes `symb` is not already on the board
	fn step_symb(&self, board: &Board, symb: Symb) -> PlayerAction {
		if board.contains(symb) {
			panic!("Called `step_symb` with a symbol that's already on the board!")
		}

		// Fill distance array with largest possible value
		let mut dist = [board.size() + 1; 11];

		// Set up initial distances
		dist[0] = 1;
		*dist.last_mut().unwrap() = 1;
		for (i, o) in board.iter().enumerate() {
			if let Some((s, _)) = o {
				if s.is_op() {
					dist[i] = 0
				}
			}
		}

		let mut did_something = true;
		while did_something {
			did_something = false;
			for i in 1..(dist.len() - 1) {
				let l = dist[i - 1];
				let r = dist[i + 1];

				let new = (l + 1).min(r + 1);
				if new < dist[i] {
					did_something = true;
					dist[i] = new;
				}
			}
		}
		let mut max_dist = *dist.iter().max().unwrap();

		loop {
			for pos in 0..11 {
				if dist[pos] >= max_dist {
					let action = PlayerAction { symb, pos };
					if board.can_play(&action) {
						return action;
					};
				}
			}

			if max_dist == 0 {
				panic!("there is no valid move!")
			}

			max_dist -= 1;
		}
	}
}

impl MinimizerAgent for DiffuseAgent {
	fn step_min(&mut self, board: &Board) -> PlayerAction {
		let symb = [Symb::Minus, Symb::Times, Symb::Plus, Symb::Div]
			.iter()
			.filter(|x| !board.contains(**x))
			.next();

		if symb.is_none() {
			// No symbols available, play a random number
			RandomAgent {}.step_min(board)
		} else {
			self.step_symb(board, *symb.unwrap())
		}
	}
}
